﻿
using System.Diagnostics;

record Corn {
	public int Weight { get; init; }
	public float Quality { get; init; }
}

class CropFarmer {
	public CropFarmer(float cornQuality, int initialWeightInStock) {
		_cornQuality = cornQuality;
		_weightInStock = initialWeightInStock;
	}

	private float _cornQuality;
	private int _weightInStock;

	public Corn? OrderCorn(int targetWeight) {
		Console.WriteLine($"### CropFarmer: got order for {targetWeight} kg of corn ...");
		if (targetWeight <= _weightInStock) {
			Console.WriteLine($"### CropFarmer: SHIPPING {targetWeight} kg of corn ~ FROM STOCK ~ ...");
			_weightInStock -= targetWeight;
			return new Corn { Weight = targetWeight, Quality = _cornQuality };
		} else {
			Console.WriteLine($"### CropFarmer: {targetWeight} kg of corn not available, cancelling corn order ...");
			return null;
		}
	}
}

record Pork {
	public float Weight { get; init; }
}

class PigProducer {
	private const int CornWeightPerPig = 50;
	private const float PorkPerPig = 0.2f;

	public required CropFarmer FavoriteFarmer { get; init; }

	public int GetPigsNeededForPork(float targetPorkWeight) => (int) Math.Round(targetPorkWeight / PorkPerPig, MidpointRounding.ToPositiveInfinity);

	public Pork? OrderPork(int pigCount) {
		Console.WriteLine($"++ PigProducer: got order for {pigCount} pigs worth of pork ...");

		int cornNeeded = pigCount * CornWeightPerPig;

		Corn? corn = FavoriteFarmer.OrderCorn(cornNeeded);
		if (corn is null) {
			Console.WriteLine($"++ PigProducer: unable to order {cornNeeded} kg of corn, cancelling pork order ...");
			return null;
		}
		Console.WriteLine($"++ PigProducer: got {corn.Weight} kg of corn of {(int) (corn.Quality * 100)} % quality ...");

		Console.WriteLine("++ PigProducer: building pig pen ...");
		Thread.Sleep(500);

		Console.WriteLine("++ PigProducer: pen is ready, need to feed pigs ...");
		Console.WriteLine($"++ PigProducer: feeding {corn.Weight} kg of corn of {(int) (corn.Quality * 100)} % quality to pigs !!! TO USE FIRST CORN !!!");
		Thread.Sleep(1500);

		int pigsProduced = (int) (corn.Weight * corn.Quality / CornWeightPerPig);
		Console.WriteLine($"++ PigProducer: {pigsProduced} pigs produced ...");

		var producedPork = new Pork { Weight = pigsProduced * PorkPerPig };
		Console.WriteLine($"++ PigProducer: SHIPPING {producedPork.Weight} kg of pork from {pigsProduced} [NEWLY PRODUCED] pigs ...");

		return producedPork;
	}
}

record HamburgerPatty { }

class MeatProcessingPlant {
	private const float PorkPerHamburgerPatty = 0.1f;

	public required PigProducer FavoritePigProducer { get; init; }

	public IReadOnlyList<HamburgerPatty>? OrderHamburgerPatties(int count) {
		Console.WriteLine($"> MeatProcessingPlant: got order for {count} hamburger patties ...");

		float porkNeeded = count * PorkPerHamburgerPatty;

		int pigsNeeded = FavoritePigProducer.GetPigsNeededForPork(porkNeeded);

		Pork? pork = FavoritePigProducer.OrderPork(pigsNeeded);
		if (pork is null) {
			Console.WriteLine($"> MeatProcessingPlant: unable to order {pigsNeeded} pigs for {porkNeeded} kg of pork, cancelling hamburger patty order ...");
			return null;
		}
		Console.WriteLine($"> MeatProcessingPlant: got {pork.Weight} kg of pork ...");

		Console.WriteLine("> MeatProcessingPlant: hiring oompa loompas from jungle...");
		Thread.Sleep(1000);

		Console.WriteLine("> MeatProcessingPlant: got oompa loompas, pork processing can start ...");
		Console.WriteLine($"> MeatProcessingPlant: producing hamburger patties from {pork.Weight} kg of pork !!! TO USE FIRST PORK !!!");

		var pattiesToProduce = (int) (pork.Weight / PorkPerHamburgerPatty);
		var producedPatties = new List<HamburgerPatty>();
		for (int i = 0; i < pattiesToProduce; i++) {
			Thread.Sleep(200);
			Console.WriteLine($"> MeatProcessingPlant: 1 hamburger patty produced ...");
			producedPatties.Add(new HamburgerPatty());
		}

		var pattiesToShip = producedPatties;
		Console.WriteLine($"> MeatProcessingPlant: SHIPPING {pattiesToShip.Count} [NEWLY PRODUCED] hamburger patties ...");

		return pattiesToShip;
	}
}

class FastFood {
	private const int HamburgersNeeded = 7;

	public required MeatProcessingPlant HamburgerPattyProvider { get; init; }

	public void Run() {
		Console.WriteLine($"FastFood: ordering {HamburgersNeeded} hamburger patties from a MeatProcessingPlant ...");
		IReadOnlyList<HamburgerPatty>? hamburgerPatties = HamburgerPattyProvider.OrderHamburgerPatties(HamburgersNeeded);
		if (hamburgerPatties is not null) {

			Console.WriteLine("FastFood: hamburger patty order successful, building new restaurant ...");
			Thread.Sleep(3000);

			Console.WriteLine("FastFood: restaurant ready, need hamburgers for opening ...");

			Console.WriteLine("FastFood: grand opening !!! TO USE FIRST HAMBURGER PATTIES !!!");
			Console.WriteLine($"FastFood: serving {hamburgerPatties.Count} hamburgers to customers ...");
		}
	}
}

class Program {
	private static void Main(string[] args) {
		var cropFarmer = new CropFarmer(0.75f, 300);
		var pigProducer = new PigProducer { FavoriteFarmer = cropFarmer };
		var meatProcessingPlant = new MeatProcessingPlant { FavoritePigProducer = pigProducer };

		var fastFood = new FastFood { HamburgerPattyProvider = meatProcessingPlant };

		var sw = Stopwatch.StartNew();
		fastFood.Run();
		sw.Stop();

		Console.WriteLine($"Ellapsed time: {sw.Elapsed.TotalSeconds} days");
	}
}
