﻿
using System.Diagnostics;

record Corn {
	public int Weight { get; init; }
	public float Quality { get; init; }
}

class CropFarmer {
	public CropFarmer(float cornQuality, int initialWeightInStock) {
		_cornQuality = cornQuality;
		_weightInStock = initialWeightInStock;
	}

	private float _cornQuality;
	private int _weightInStock;

	public Task<Corn>? OrderCornAsync(int targetWeight) {
		Console.WriteLine($"### CropFarmer: got order for {targetWeight} kg of corn ...");
		if (targetWeight <= _weightInStock) {
			Console.WriteLine($"### CropFarmer: SHIPPING {targetWeight} kg of corn ~~ FROM STOCK ~~ ...");
			_weightInStock -= targetWeight;
			return Task.FromResult(new Corn { Weight = targetWeight, Quality = _cornQuality });
		} else {
			Console.WriteLine($"### CropFarmer: {targetWeight} kg of corn not available, cancelling corn order ...");
			return null;
		}
	}
}

record Pork {
	public float Weight { get; init; }
}

class PigProducer {
	private const int CornWeightPerPig = 50;
	private const float PorkPerPig = 0.2f;

	public required CropFarmer FavoriteFarmer { get; init; }

	public int GetPigsNeededForPork(float targetPorkWeight) => (int) Math.Round(targetPorkWeight / PorkPerPig, MidpointRounding.ToPositiveInfinity);

	public Task<Pork>? OrderPorkAsync(int pigCount) {
		Console.WriteLine($"++ PigProducer: got order for {pigCount} pigs worth of pork ...");

		int cornNeeded = pigCount * CornWeightPerPig;

		Task<Corn>? futureCorn = FavoriteFarmer.OrderCornAsync(cornNeeded);
		if (futureCorn is null) {
			Console.WriteLine($"++ PigProducer: unable to order {cornNeeded} kg of corn, cancelling pork order ...");
			return null;
		}
		Console.WriteLine($"++ PigProducer: corn order successful, order state: futureCorn.IsCompleted -> {futureCorn.IsCompleted} ...");

		ThreadPool.QueueUserWorkItem(_ => {
			Console.WriteLine("++ PigProducer: building pig pen ...");
			Thread.Sleep(500);

			Console.WriteLine($"++ PigProducer: pen is ready, need to feed pigs ... Is futureCorn ready? .IsCompleted -> {futureCorn.IsCompleted}");
			Corn corn = futureCorn.Result;
			Console.WriteLine($"++ PigProducer: got {corn.Weight} kg of corn of {(int) (corn.Quality * 100)} % quality, feeding to pigs !!! TO USE FIRST CORN !!!");
			Thread.Sleep(1500);

			int pigsProduced = (int) (corn.Weight * corn.Quality / CornWeightPerPig);
			Console.WriteLine($"++ PigProducer: {pigsProduced} pigs FOR FUTURE ORDERS produced ...");
		});

		// Use pork in stock:
		var producedPork = new Pork { Weight = pigCount * PorkPerPig };
		Console.WriteLine($"++ PigProducer: SHIPPING {producedPork.Weight} kg of pork from {pigCount} pigs ~~ FROM STOCK ~~ ...");

		return Task.FromResult(producedPork);
	}
}

record HamburgerPatty { }

class MeatProcessingPlant {
	private const float PorkPerHamburgerPatty = 0.1f;

	public required PigProducer FavoritePigProducer { get; init; }

	public Task<IReadOnlyList<HamburgerPatty>>? OrderHamburgerPattiesAsync(int count) {
		Console.WriteLine($"> MeatProcessingPlant: got order for {count} hamburger patties ...");

		float porkNeeded = count * PorkPerHamburgerPatty;

		int pigsNeeded = FavoritePigProducer.GetPigsNeededForPork(porkNeeded);

		Task<Pork>? futurePork = FavoritePigProducer.OrderPorkAsync(pigsNeeded);
		if (futurePork is null) {
			Console.WriteLine($"> MeatProcessingPlant: unable to order {pigsNeeded} pigs for {porkNeeded} kg of pork, cancelling hamburger patty order ...");
			return null;
		}
		Console.WriteLine($"> MeatProcessingPlant: pork order successful, order state: futurePork.IsCompleted -> {futurePork.IsCompleted} ...");

		ThreadPool.QueueUserWorkItem(_ => {
			Console.WriteLine("> MeatProcessingPlant: hiring oompa loompas from jungle ...");
			Thread.Sleep(1000);

			Console.WriteLine($"> MeatProcessingPlant: got oompa loompas, pork processing can start ... Is futurePork ready? .IsCompleted -> {futurePork.IsCompleted}");
			Pork pork = futurePork.Result;
			Console.WriteLine($"> MeatProcessingPlant: got {pork.Weight} kg of pork, producing hamburger patties !!! TO USE FIRST PORK !!!");

			var pattiesToProduce = (int) (pork.Weight / PorkPerHamburgerPatty);
			var producedPatties = new List<HamburgerPatty>();
			for (int i = 0; i < pattiesToProduce; i++) {
				Thread.Sleep(200);
				Console.WriteLine($"> MeatProcessingPlant: 1 hamburger patty FOR FUTURE ORDERS produced ...");
				producedPatties.Add(new HamburgerPatty());
			}
		});

		// Use hamburger patties in stock:
		var pattiesToShip = new List<HamburgerPatty> { new HamburgerPatty(), new HamburgerPatty(), new HamburgerPatty() };
		Console.WriteLine($"> MeatProcessingPlant: SHIPPING {pattiesToShip.Count} hamburger patties ~~ FROM STOCK ~~ ...");

		return Task.FromResult<IReadOnlyList<HamburgerPatty>>(pattiesToShip);
	}
}

class FastFood {
	private const int HamburgersNeeded = 7;

	public required MeatProcessingPlant HamburgerPattyProvider { get; init; }

	public void Run() {
		Console.WriteLine($"FastFood: ordering {HamburgersNeeded} hamburger patties from a MeatProcessingPlant ...");
		Task<IReadOnlyList<HamburgerPatty>>? futureHamburgerPatties = HamburgerPattyProvider.OrderHamburgerPattiesAsync(HamburgersNeeded);
		if (futureHamburgerPatties is not null) {
			Console.WriteLine($"FastFood: hamburger patty order successful, order state: futureHamburgerPatties.IsCompleted -> {futureHamburgerPatties.IsCompleted} ...");
			Console.WriteLine("FastFood: building new restaurant ...");
			Thread.Sleep(3000);

			Console.WriteLine($"FastFood: restaurant ready, need hamburgers for opening ... Are futureHamburgerPatties ready? .IsCompleted -> {futureHamburgerPatties.IsCompleted}");
			IReadOnlyList<HamburgerPatty> hamburgers = futureHamburgerPatties.Result;

			Console.WriteLine("FastFood: grand opening !!! TO USE FIRST HAMBURGER PATTIES !!!");
			Console.WriteLine($"FastFood: serving {hamburgers.Count} hamburgers to customers ...");
		}
	}
}

class Program {
	private static void Main(string[] args) {
		var cropFarmer = new CropFarmer(0.75f, 300);
		var pigProducer = new PigProducer { FavoriteFarmer = cropFarmer };
		var meatProcessingPlant = new MeatProcessingPlant { FavoritePigProducer = pigProducer };

		var fastFood = new FastFood { HamburgerPattyProvider = meatProcessingPlant };

		var sw = Stopwatch.StartNew();
		fastFood.Run();
		sw.Stop();

		Console.WriteLine($"Ellapsed time: {sw.Elapsed.TotalSeconds} days");
	}
}

